@@ -24,6 +24,7 @@ from .caps import (
 )
 from .ffi import sys_exit as _sys_exit
 from .file_caps import FileCaps
+from .fsid import getfsgid, getfsuid, setfsgid, setfsuid
 from .misc import (
     MCEKillPolicy,
     TimingMethod,
@@ -71,6 +72,10 @@ __all__ = (
     "cap_effective",
     "cap_ambient",
     "securebits",
+    "getfsuid",
+    "getfsgid",
+    "setfsuid",
+    "setfsgid",
     "FileCaps",
     "MCEKillPolicy",
     "TimingMethod",
@@ -96,4 +101,4 @@ __all__ = (
     "_sys_exit",
 )
 
-__version__ = "0.1.1"
+__version__ = "0.1.3"
