@@ -1,10 +1,11 @@
+# pylint: disable=too-many-lines
 import contextlib
 import ctypes
 import dataclasses
 import enum
 import errno
-import os
 import re
+import string
 import warnings
 from typing import Callable, Iterable, Iterator, List, Optional, Set, Tuple, cast
 
@@ -93,18 +94,21 @@ class Cap(enum.Enum):
         Roughly equivalent to ``cap_from_name()`` in libcap. Names are matched case-insensitively,
         but they must include a ``cap_`` prefix (also case-insensitive; ``CAP_`` and ``Cap_`` are
         valid too).
-
         """
 
         upper_name = name.upper()
 
-        if upper_name.startswith("CAP_"):
+        if (
+            upper_name.startswith("CAP_")
+            and len(upper_name) > 4
+            and upper_name[4] in string.ascii_uppercase
+        ):
             try:
                 return cast(Cap, getattr(cls, upper_name[4:]))
             except AttributeError:
                 pass
 
-        raise ValueError("Unknown capability {!r}".format(name))
+        raise ValueError(f"Unknown capability {name!r}")
 
     def is_supported(self) -> bool:
         """
@@ -117,10 +121,29 @@ class Cap(enum.Enum):
         """
         Returns the set of capabilities supported by the running kernel.
         """
-        return {cap for cap in cls if cap.is_supported()}
+
+        # Do a binary search
+
+        # CAP_MAC_ADMIN was added in kernel 2.6.25, so this will give us support for everything
+        # since then (if supported by the libc)
+        low = Cap.MAC_ADMIN.value
+        high = _LAST_CAP.value
+
+        while low != high:
+            # This basically does `mid = ceil((low + high) / 2)`.
+            # If we don't do ceiling division, the way binary search works, we'll get stuck at
+            # `high = low + 1` forever.
+            mid = (low + high + 1) >> 1
+
+            if cls(mid).is_supported():
+                low = mid
+            else:
+                high = mid - 1
+
+        return {cls(i) for i in range(low + 1)}
 
 
-_LAST_CAP = max(Cap, key=lambda cap: cap.value)  # type: ignore
+_LAST_CAP = max(Cap, key=lambda cap: cap.value)
 _ALL_CAPS_SET = set(Cap)
 
 
@@ -178,11 +201,13 @@ class _CapabilitySet:
     def drop(self, *drop_caps: Cap) -> None:
         if self._name == "bounding":
             for cap in drop_caps:
-                capbset_drop(cap)
+                if capbset_read(cap):
+                    capbset_drop(cap)
 
         elif self._name == "ambient":
             for cap in drop_caps:
-                cap_ambient_lower(cap)
+                if cap_ambient_is_set(cap):
+                    cap_ambient_lower(cap)
 
         else:
             state = CapState.get_current()
@@ -225,7 +250,9 @@ class _CapabilitySet:
         if not limit_caps:
             self.clear()
         else:
-            self.drop(*(set(Cap) - set(limit_caps)))
+            self.drop(*(_ALL_CAPS_SET - set(limit_caps)))
+
+            self._clear_extra()
 
     def clear(self) -> None:
         if self._name == "ambient":
@@ -233,13 +260,59 @@ class _CapabilitySet:
         else:
             self.drop(*Cap)
 
+            if self._name == "bounding":
+                self._clear_extra()
+
+    def replace(self, *caps: Cap) -> None:
+        if not caps:
+            self.clear()
+            return
+
+        if self._name == "bounding":
+            for cap in Cap:
+                is_raised = capbset_read(cap)
+                want_raised = cap in caps
+
+                if want_raised and not is_raised:
+                    raise ValueError("Cannot add bounding capabilities")
+                elif is_raised and not want_raised:
+                    capbset_drop(cap)
+
+            self._clear_extra()
+
+        elif self._name == "ambient":
+            for cap in Cap:
+                is_raised = cap_ambient_is_set(cap)
+                want_raised = cap in caps
+
+                if want_raised and not is_raised:
+                    cap_ambient_raise(cap)
+                elif is_raised and not want_raised:
+                    cap_ambient_lower(cap)
+
+            self._clear_extra()
+
+        else:
+            state = CapState.get_current()
+            setattr(state, self._name, caps)
+            state.set_current()
+
+    def _clear_extra(self) -> None:
+        if self._name == "bounding":
+            for i in range(_LAST_CAP.value + 1, _LAST_KERNEL_CAP + 1):
+                if ffi.prctl(ffi.PR_CAPBSET_READ, i, 0, 0, 0):
+                    ffi.prctl(ffi.PR_CAPBSET_DROP, i, 0, 0, 0)
+
+        elif self._name == "ambient":
+            for i in range(_LAST_CAP.value + 1, _LAST_KERNEL_CAP + 1):
+                ffi.prctl(ffi.PR_CAP_AMBIENT_LOWER, i, 0, 0, 0)
+
     def __repr__(self) -> str:
-        return "<{} capability set: {}>".format(self._name.title(), self)
+        return f"<{self._name.title()} capability set: {self}>"
 
     def __str__(self) -> str:
         return "=" + ",".join(
-            "cap_" + cap.name.lower()
-            for cap in sorted(self.probe(), key=lambda cap: cap.value)  # type: ignore
+            "cap_" + cap.name.lower() for cap in sorted(self.probe(), key=lambda cap: cap.value)
         )
 
 
@@ -328,6 +401,9 @@ class CapState:
         """
         Set the capability state of the current thread to the capability set represented by this
         object.
+
+        Note: If the capability sets stored in this object contain capabilities that the running
+        kernel does not support, the kernel will silently ignore them!
         """
 
         header = _CapUserHeader(
@@ -377,9 +453,9 @@ def _capset_from_bitmask(bitmask: int) -> Set[Cap]:
                 res.add(Cap(i))
             else:
                 warnings.warn(
-                    "Unrecognized capability (number {}) found in capability set. This may result "
-                    "in strange behavior. Are you using an old version of pyprctl on a newer "
-                    "kernel?".format(i),
+                    f"Unrecognized capability (number {i}) found in capability set. This may "
+                    "result in strange behavior. Are you using an old version of pyprctl on a "
+                    "newer kernel?",
                     RuntimeWarning,
                 )
 
@@ -467,7 +543,7 @@ def _capstate_from_text(text: str) -> Tuple[Set[Cap], Set[Cap], Set[Cap]]:
                     permitted.difference_update(caps)
 
             else:
-                raise ValueError("Invalid character {!r} in capability set clause".format(ch))
+                raise ValueError(f"Invalid character {ch!r} in capability set clause")
 
             last_ch = ch
 
@@ -483,8 +559,7 @@ def _capstate_to_text(*, effective: Set[Cap], inheritable: Set[Cap], permitted:
             return "" if prefix_ch == "=" else "all"
 
         return ",".join(
-            "cap_" + cap.name.lower()
-            for cap in sorted(caps, key=lambda cap: cap.value)  # type: ignore
+            "cap_" + cap.name.lower() for cap in sorted(caps, key=lambda cap: cap.value)
         )
 
     orig_effective = effective
@@ -696,13 +771,13 @@ class _SecurebitsAccessor:  # pylint: disable=too-few-public-methods
     }
 
     def __repr__(self) -> str:
-        return "<Securebits: {}>".format(self)
+        return f"<Securebits: {self}>"
 
     def __str__(self) -> str:
         cur_secbits = get_securebits()
 
         return ", ".join(
-            "secure-{}: {} ({})".format(
+            "secure-{}: {} ({})".format(  # pylint: disable=consider-using-f-string
                 secbit.name.lower().replace("_", "-"),
                 "yes" if secbit in cur_secbits else "no",
                 "locked" if lock_secbit in cur_secbits else "unlocked",
@@ -737,6 +812,8 @@ def capbset_drop(cap: Cap) -> None:
     Remove the given capability from the current thread's bounding capability set.
 
     (This requires the CAP_SETPCAP capability.)
+
+    This function will fail with ``EINVAL`` if the kernel does not support the specified capability.
     """
     ffi.prctl(ffi.PR_CAPBSET_DROP, cap.value, 0, 0, 0)
 
@@ -755,6 +832,8 @@ def cap_ambient_raise(cap: Cap) -> None:
 
     (The capability must already be present in the thread's permitted set and and the thread's
     inheritable set, and the SECBIT_NO_CAP_AMBIENT_RAISE securebit must not be set.)
+
+    This function will fail with ``EINVAL`` if the kernel does not support the specified capability.
     """
     ffi.prctl(ffi.PR_CAP_AMBIENT, ffi.PR_CAP_AMBIENT_RAISE, cap.value, 0, 0)
 
@@ -762,6 +841,8 @@ def cap_ambient_raise(cap: Cap) -> None:
 def cap_ambient_lower(cap: Cap) -> None:
     """
     Lower the given capability in the current thread's ambient set.
+
+    This function will fail with ``EINVAL`` if the kernel does not support the specified capability.
     """
     ffi.prctl(ffi.PR_CAP_AMBIENT, ffi.PR_CAP_AMBIENT_LOWER, cap.value, 0, 0)
 
@@ -837,20 +918,18 @@ def cap_set_ids(
       will be restored to its original contents. By default, this function mimics ``libcap`` and
       empties the effective capability set before returning.
 
-    Note: If an error occurs while this function is attempting to change the thread's
-    UID/GID/supplementary groups, this function will still attempt to set the effective capability
-    set as specified by ``preserve_effective_caps``. However, this may fail. Programs that plan on
-    performing further actions if this function returns an error should first check the thread's
-    capability set and ensure that the correct capabilities are raised/lowered.
+    Note: If this function fails and raises an ``OSError``, the thread's UIDs, GIDs, supplementary
+    groups, and capability sets are in an unknown and possibly inconsistent state. This is EXTREMELY
+    DANGEROUS! If you are unable to revert the changes, abort as soon as possible.
     """
 
     # We do type checks up front to avoid errors in the middle of changing the UIDs/GIDs.
 
     if uid is not None and not isinstance(uid, int):
-        raise TypeError("Invalid type {!r} for 'uid' argument".format(uid.__class__.__name__))
+        raise TypeError(f"Invalid type {uid.__class__.__name__!r} for 'uid' argument")
 
     if gid is not None and not isinstance(gid, int):
-        raise TypeError("Invalid type {!r} for 'gid' argument".format(gid.__class__.__name__))
+        raise TypeError(f"Invalid type {gid.__class__.__name__!r} for 'gid' argument")
 
     if groups is not None:
         groups = list(groups)
@@ -858,9 +937,7 @@ def cap_set_ids(
         for supp_gid in groups:
             if not isinstance(supp_gid, int):
                 raise TypeError(
-                    "Invalid type {!r} for value in 'groups' list".format(
-                        supp_gid.__class__.__name__
-                    )
+                    f"Invalid type {supp_gid.__class__.__name__!r} for value in 'groups' list"
                 )
 
     if uid is None and gid is None and groups is None:
@@ -892,7 +969,7 @@ def cap_set_ids(
                 ffi.sys_setresgid(gid, gid, gid)
 
             if groups is not None:
-                os.setgroups(groups)
+                ffi.sys_setgroups(groups)
 
             if uid is not None:
                 ffi.sys_setresuid(uid, uid, uid)
@@ -928,7 +1005,6 @@ def scoped_effective_caps(effective: Iterable[Cap]) -> Iterator[None]:
         ambient, and bounding) in the body of the context manager *will* be preserved. (Be careful
         not to remove any of the capabilities present in the original effective set from the
         permitted set, or this function may fail to revert to the original effective set.)
-
     """
 
     effective = set(effective)
@@ -950,3 +1026,29 @@ def scoped_effective_caps(effective: Iterable[Cap]) -> Iterator[None]:
         capstate = CapState.get_current()
         capstate.effective = orig_effective
         capstate.set_current()
+
+
+# This is only set to anything other than `_LAST_CAP.value` if the kernel supports capabilities that
+# this library does not (in which case it is set to the number of the final capability supported by
+# the kernel).
+#
+# This is then used to clear the "extra" capabilities not supported by the kernel when certain
+# operations are performed with the high-level APIs.
+_LAST_KERNEL_CAP = _LAST_CAP.value
+
+while True:
+    try:
+        ffi.prctl(ffi.PR_CAPBSET_READ, _LAST_KERNEL_CAP + 1, 0, 0, 0)
+    except OSError as _ex:
+        assert _ex.errno == errno.EINVAL
+        break
+    else:
+        _LAST_KERNEL_CAP += 1
+
+if _LAST_KERNEL_CAP > _LAST_CAP.value:
+    warnings.warn(
+        f"Unrecognized capabilities (numbers {_LAST_CAP.value + 1} to {_LAST_KERNEL_CAP}) detected."
+        " This may result in strange and possibly dangerous behavior. Are you using an old version "
+        "of pyprctl on a newer kernel?",
+        RuntimeWarning,
+    )
